using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Runtime.InteropServices;
using System.Text;
using System.Windows.Forms;

namespace USBDevCtrl
{
	public partial class DevCtrlForm : Form
	{
        //Application global variables 
        HIDDevice Device = new HIDDevice();

        // Vendor and product IDs to look for.
        const uint VendorID = 0x04D8;
        const uint ProductID = 0x4541;

		int motorSpeed = 0;
		byte solenoidVoltage = 0;
		bool solenoidFiring = false;
		int servoPosition = 0;
		byte PORTA, PORTB, PORTC;
		byte devMotorSpeed, devServoPosition;
        
        public DevCtrlForm()
		{
			InitializeComponent();

			System.Guid HidGuid = System.Guid.NewGuid();
			HIDDevice.HidD_GetHidGuid(ref HidGuid);
			Device.RegisterForDeviceNotifications(this.Handle, HidGuid);

			backgroundWorker1.RunWorkerAsync();
		}

		protected override void WndProc(ref Message m)
		{
			if (m.Msg == HIDDevice.WM_DEVICECHANGE)
			{
				int wParam = m.WParam.ToInt32();

				if (wParam == HIDDevice.DBT_DEVICEARRIVAL)
				{
					int devType = Marshal.ReadInt32(m.LParam, 4);
					if (devType == HIDDevice.DBT_DEVTYP_DEVICEINTERFACE)
					{
						HIDDevice.DEV_BROADCAST_DEVICEINTERFACE devIF;
						devIF = (HIDDevice.DEV_BROADCAST_DEVICEINTERFACE)Marshal.PtrToStructure(m.LParam, typeof(HIDDevice.DEV_BROADCAST_DEVICEINTERFACE));
						// TBD: See if it is the device we are interested in
						// For now, set DeviceDetected to false on any
						// interface change.  Slight performance penalty,
						// but will ensure we don't try talking to a device
						// that isn't there.
						// Ensure the device is available.
						Device.DeviceDetected = false;
					}
				}
				else if (wParam == HIDDevice.DBT_DEVICEREMOVECOMPLETE)
				{
					int devType = Marshal.ReadInt32(m.LParam, 4);
					// A device was removed
					if (devType == HIDDevice.DBT_DEVTYP_DEVICEINTERFACE)
					{
						Device.DeviceDetected = false;
					}
				}
			}
			base.WndProc(ref m);
		}

		// Send a GET_FEATURE to the device
        bool GetFeature(byte reportID)
        {
            // Ensure the device is available.
            if (!Device.DeviceDetected)
            {
                if (!Device.FindHID(VendorID, ProductID))
                    return false;
            }

            Array.Clear(Device.FeatureReport, 0, Device.FeatureReport.Length);
            Device.FeatureReport[0] = reportID;
            return HIDDevice.HidD_GetFeature(Device.HidDevice, Device.FeatureReport, Device.Caps.FeatureReportByteLength);
        }

        // Send a SET_FEATURE to the device
        bool SetFeature(byte reportID)
        {
            // Ensure the device is available.
            if (!Device.DeviceDetected)
            {
                if (!Device.FindHID(VendorID, ProductID))
                    return false;
            }

            // Ensure the feature report is clear before sending.
            Array.Clear(Device.FeatureReport, 0, Device.FeatureReport.Length);
            Device.FeatureReport[0] = reportID;
            // Accumulate bits from the checkboxes (device will use them
            // to do something like turn on/off LEDs).
            Device.FeatureReport[1] = 0;
            if (checkBox1.Checked)
                Device.FeatureReport[1] |= 0x01;
            if (checkBox2.Checked)
                Device.FeatureReport[1] |= 0x02;
            if (checkBox3.Checked)
                Device.FeatureReport[1] |= 0x04;
            if (checkBox4.Checked)
                Device.FeatureReport[1] |= 0x08;
            if (checkBox5.Checked)
                Device.FeatureReport[1] |= 0x10;
            if (checkBox6.Checked)
                Device.FeatureReport[1] |= 0x20;
            if (checkBox7.Checked)
                Device.FeatureReport[1] |= 0x40;
            if (checkBox8.Checked)
                Device.FeatureReport[1] |= 0x80;
			int mtrSpd = (enableMotorCheckBox.Checked ? motorSpeed : 0);
			Device.FeatureReport[2] = (byte)(mtrSpd + 128); // Data byte 1
			Device.FeatureReport[3] = (byte)(servoPosition); // Data byte 2
			Device.FeatureReport[4] = (byte)(solenoidFiring ? 0x01 : 0); // Data byte 3
			Device.FeatureReport[5] = 0x42; // Data byte 4
			bool Result = HIDDevice.HidD_SetFeature(Device.HidDevice, Device.FeatureReport,
                Device.Caps.FeatureReportByteLength);
            return Result;
        }

        // Sit in a loop gathering input reports from the device.
		private void backgroundWorker1_DoWork(object sender, DoWorkEventArgs e)
		{
			// This method will run on a thread other than the UI thread.
			// Be sure not to manipulate any Windows Forms controls created
			// on the UI thread from this method.
			// Get the BackgroundWorker that raised this event.
			BackgroundWorker worker = sender as BackgroundWorker;

			// Keep working until told to stop
			while (!worker.CancellationPending)
			{
				// Wait a bit
				System.Threading.Thread.Sleep(100);

				// Ask the device for information
				if (!GetFeature(0))
				{
					// Failed to get info - take a longer sleep this time
					System.Threading.Thread.Sleep(1000);
					continue;
				}

				// Put the info into local variables
				PORTA = Device.FeatureReport[1];
				PORTB = Device.FeatureReport[2];
				PORTC = Device.FeatureReport[3];
				solenoidVoltage = Device.FeatureReport[4];
				devMotorSpeed = Device.FeatureReport[5];
				devServoPosition = Device.FeatureReport[6];

				// Update the GUI
				worker.ReportProgress(0);
			}
		}

		// Update UI based on data that has come from the USB device
		private void backgroundWorker1_ProgressChanged(object sender, ProgressChangedEventArgs e)
		{
			// Display information as appropriate
			solenoidVoltageBar.Value = solenoidVoltage;
		}

		private void DevCtrlForm_FormClosing(object sender, FormClosingEventArgs e)
		{
			// If we are exiting the program, then no need to keep the background
			// worker alive.
			backgroundWorker1.CancelAsync();
		}

		private void motorTrackBar_ValueChanged(object sender, EventArgs e)
		{
			int val = motorTrackBar.Value;
			motorSpeed = val;
			SetFeature(0);
		}

		private void fireSolenoidButton_Click(object sender, EventArgs e)
		{
			// SetFeature(0);
		}

		private void servoTrackBar_ValueChanged(object sender, EventArgs e)
		{
			servoPosition = servoTrackBar.Value + 128;
			SetFeature(0);
		}

		private void enableMotorCheckBox_Click(object sender, EventArgs e)
		{
			// Enable/disable motor based on checkbox value
			SetFeature(0);
		}

		private void checkBox1_Click(object sender, EventArgs e)
		{
			SetFeature(0);
		}

		private void fireSolenoidButton_MouseDown(object sender, MouseEventArgs e)
		{
			solenoidFiring = true;
			SetFeature(0);
		}

		private void fireSolenoidButton_MouseUp(object sender, MouseEventArgs e)
		{
			solenoidFiring = false;
			SetFeature(0);
		}
	}
}